@@ -1,19 +1,21 @@
 from ast_monitor.model import AST
 from ast_monitor.classes import SensorData, Interval
+from ast_monitor.digital_twin import DigitalTwin
 from ast_monitor.gps_sensor import GpsSensor
 from ast_monitor.hr_sensor import HrSensor
 from ast_monitor.simulation import Simulation
 from ast_monitor.mainwindow import Ui_MainWindow
 
 __all__ = [
-    'AST',
-    'SensorData',
-    'Interval',
-    'GpsSensor',
-    'HrSensor',
-    'Simulation',
-    'Ui_MainWindow'
+    AST,
+    SensorData,
+    Interval,
+    DigitalTwin,
+    GpsSensor,
+    HrSensor,
+    Simulation,
+    Ui_MainWindow
 ]
 
 __project__ = 'ast_monitor'
-__version__ = '0.1.5'
+__version__ = '0.2.0'
