@@ -14,14 +14,16 @@ from PyQt5 import (
     QtWebEngineWidgets
 )
 from PyQt5.QtCore import pyqtSlot, Qt, QTimer
-from PyQt5.QtWidgets import QMessageBox
+from PyQt5.QtWidgets import QLabel, QMessageBox
+from threading import Thread
 
 from ast_monitor.classes import Interval, SensorData
+from ast_monitor.digital_twin import DigitalTwin
 from ast_monitor.mainwindow import Ui_MainWindow
 from ast_monitor.send_data import SendData
 
 
-TICK_TIME = 2**6
+TICK_TIME = 1000
 
 
 class AST(QtWidgets.QMainWindow, Ui_MainWindow):
@@ -33,7 +35,6 @@ class AST(QtWidgets.QMainWindow, Ui_MainWindow):
         gps_path (str):
             path to file for storing GPS data
     """
-
     def __init__(self, hr_data_path: str, gps_data_path: str) -> None:
         """
         Initialisation method for AST class.\n
@@ -51,8 +52,18 @@ class AST(QtWidgets.QMainWindow, Ui_MainWindow):
 
         # Structures for storing data and intervals.
         self.time_series = []
+        self.planned_intervals = []
         self.intervals = []
 
+        # Current interval data.
+        self.current_interval = 0
+        self.is_speed_phase = False
+        self.interval_average_heart_rate = 0
+        self.interval_checkpoints = 0
+
+        #
+        self.training_timer = QTimer()
+
     def initialize_GUI(self) -> None:
         """
         Initialization of the AST-Monitor GUI.
@@ -73,11 +84,16 @@ class AST(QtWidgets.QMainWindow, Ui_MainWindow):
         self.view.setUrl(QtCore.QUrl.fromLocalFile(file))
         self.vb_map.addWidget(self.view)
 
-        # Show HR in real time.
+        # Show data in real time.
         self.timer = QTimer()
         self.timer.timeout.connect(self.update_data)
         self.timer.start(1000)
 
+        # Show interval data in real time.
+        self.interval_timer = QTimer()
+        self.interval_timer.timeout.connect(self.update_interval_data)
+        self.interval_timer.start(1000)
+
         # Shutdown the computer.
         self.btn_shutdown.clicked.connect(self.shutdown_now)
 
@@ -87,6 +103,12 @@ class AST(QtWidgets.QMainWindow, Ui_MainWindow):
         self.tracker.timeout.connect(self.tick)
         self.do_reset()
 
+        # Interval tick.
+        self.interval_tracker = QTimer()
+        self.interval_tracker.setInterval(TICK_TIME)
+        self.interval_tracker.timeout.connect(self.interval_tick)
+        self.do_interval_reset()
+
         # Menu buttons.
         self.action_about_program.triggered.connect(self.about)
         self.action_license.triggered.connect(self.license)
@@ -102,11 +124,11 @@ class AST(QtWidgets.QMainWindow, Ui_MainWindow):
         self.btn_move_right.clicked.connect(self.menu_right_move)
 
         # Interval buttons.
-        self.btn_start_interval.clicked.connect(self.start_interval)
-        self.btn_end_interval.clicked.connect(self.end_interval)
+        self.interval_flag = False
 
         # Training button.
         self.btn_load_training.clicked.connect(self.load_training)
+        self.btn_start_training.clicked.connect(self.start_training)
 
     def start_tracker(self) -> None:
         """
@@ -117,11 +139,6 @@ class AST(QtWidgets.QMainWindow, Ui_MainWindow):
 
         self.widget_start_stop.setCurrentIndex(1)
 
-        # Sending data in real time.
-        self.timer_data = QTimer()
-        self.timer_data.timeout.connect(self.send_data)
-        self.timer_data.start(10000)
-
         # Show simple notification that workout just started.
         self._feedback = TextFeedback(text='Workout started!')
         self._feedback.show(
@@ -152,13 +169,27 @@ class AST(QtWidgets.QMainWindow, Ui_MainWindow):
         """
         Rendering all the necessary data.
         """
-        self.update_hr()
-        self.update_speed()
+        self.update_current_heart_rate(label=self.lbl_heart_rate)
+        self.update_speed(label=self.lbl_speed)
 
         # Build_time_series in case tracker is enabled.
         if self.tracking_flag:
-            self.build_time_series()
-            self.update_distance()
+            self.build_time_series(series=self.time_series)
+            self.update_distance(label=self.lbl_distance)
+
+    @pyqtSlot()
+    def update_interval_data(self) -> None:
+        """
+        Rendering all the necessary interval data.
+        """
+        # Build_time_series in case tracker is enabled.
+        if self.interval_flag:
+            self.update_current_heart_rate(label=self.lbl_interval_heart_rate)
+            self.update_average_heart_rate(
+                self.interval_average_heart_rate,
+                self.return_current_hr(),
+                label=self.lbl_interval_average_heart_rate
+            )
 
     @pyqtSlot()
     def send_data(self) -> None:
@@ -175,6 +206,13 @@ class AST(QtWidgets.QMainWindow, Ui_MainWindow):
         self.track_time = 0
         self.display_stopwatch()
 
+    @pyqtSlot()
+    def do_interval_reset(self) -> None:
+        """
+        Reset of the interval.
+        """
+        self.interval_time = 0
+
     def tick(self) -> None:
         """
         Incrementing time and displaying it on the stopwatch on each tick.
@@ -182,12 +220,35 @@ class AST(QtWidgets.QMainWindow, Ui_MainWindow):
         self.track_time += TICK_TIME / 1000
         self.display_stopwatch()
 
-    def display_stopwatch(self) -> None:
+    def interval_tick(self) -> None:
         """
-        Displaying the stopwatch in GUI.
+        Incrementing time and displaying it on
+        the interval stopwatch on each tick.
         """
-        seconds = int(self.track_time % 60)
-        minutes = int(self.track_time / 60)
+        self.interval_time += TICK_TIME / 1000
+        self.display_interval_stopwatch()
+
+        try:
+            if self.digital_twin:
+                self.digital_twin.current_heart_rate = self.return_current_hr()
+                self.digital_twin.current_duration = self.interval_time
+                self.lbl_interval_proposed_heart_rate.setText(
+                    str(self.digital_twin.predicted_heart_rate)
+                )
+        except Exception:
+            pass
+
+    def convert_time_to_hours_minutes_seconds(self, time: int) -> str:
+        """
+        Converting time in seconds to HH:MM:SS format.\n
+        Args:
+            time (int):
+                time in seconds
+        Returns:
+            str: time in HH:MM:SS format
+        """
+        seconds = int(time % 60)
+        minutes = int(time / 60)
         hours = int(minutes / 60)
 
         time = (
@@ -198,47 +259,89 @@ class AST(QtWidgets.QMainWindow, Ui_MainWindow):
             (str(seconds) if seconds > 9 else '0' + str(seconds))
         )
 
+        return time
+
+    def display_stopwatch(self) -> None:
+        """
+        Displaying the stopwatch in the GUI.
+        """
+        time = self.convert_time_to_hours_minutes_seconds(self.track_time)
         self.lbl_watch.setText(time)
 
-    def update_distance(self) -> None:
+    def display_interval_stopwatch(self) -> None:
         """
-        Calculating and displaying distance.
+        Displaying the stopwatch in the GUI.
         """
-        dist = self.calculate_distance()
+        time = self.convert_time_to_hours_minutes_seconds(self.interval_time)
+        self.lbl_interval_watch.setText(time)
+
+    def update_current_heart_rate(self, label: QLabel) -> None:
+        """
+        Calculating and displaying current heart rate.\n
+        Args:
+            label (QLabel):
+                label to be updated
+        """
+        hr = self.return_current_hr()
+        label.setText(str(hr))
+
+    def update_average_heart_rate(
+        self,
+        previous_heart_rate: int,
+        current_heart_rate: int,
+        label: QLabel
+    ) -> None:
+        """
+        Calculating and displaying average heart rate
+        with a help of a recursive function.\n
+        Args:
+            previous_heart_rate (int):
+                previous average heart rate
+            current_heart_rate (int):
+                current heart rate
+            label (QLabel):
+                label to be updated
+        """
+        if previous_heart_rate == -1:
+            return
+        elif previous_heart_rate == 0:
+            average_heart_rate = current_heart_rate
+        else:
+            n = self.interval_checkpoints
+            avg_hr = previous_heart_rate
+            hr = current_heart_rate
+            average_heart_rate = int((1 / n) * ((n - 1) * avg_hr + hr))
+
+        label.setText(str(average_heart_rate))
+        self.interval_average_heart_rate = average_heart_rate
+        self.interval_checkpoints += 1
+
+    def update_speed(self, label: QLabel) -> None:
+        """
+        Calculating and displaying current speed.\n
+        Args:
+            label (QLabel):
+                label to be updated
+        """
+        speed = self.calculate_speed()
+        label.setText(f'{speed} km/h')
+
+    def update_distance(self, label: QLabel) -> None:
+        """
+        Calculating and displaying distance.\n
+        Args:
+            label (QLabel):
+                label to be updated
+        """
+        dist = self.calculate_distance(self.time_series)
 
         # If there is no distance made yet, 0.00 km is displayed.
         if not dist:
-            self.lbl_distance.setText('0.00 km')
+            label.setText('0.00 km')
         else:
             rounded_dist = round(dist, 2)
             dist_str = '{:.2f}'.format(rounded_dist / 1000)
-            self.lbl_distance.setText(dist_str + ' km')
-
-    def update_speed(self) -> None:
-        """
-        Calculating and updating speed.
-        """
-        speed = self.calculate_speed()
-        self.lbl_speed.setText(f'{speed} km/h')
-
-    def update_ascent(self) -> None:
-        """
-        Calculating and updating ascent.
-        """
-        pass
-
-    def update_hr(self) -> None:
-        """
-        Calculating and displaying average heart rate.
-        """
-        hr = self.return_current_hr()
-        self.lbl_heart_rate.setText(str(hr))
-
-    def update_interval_hr(self) -> None:
-        """
-        Calculating and displaying interval heart rate.
-        """
-        pass
+            label.setText(dist_str + ' km')
 
     def shutdown_now(self) -> None:
         """
@@ -265,16 +368,19 @@ class AST(QtWidgets.QMainWindow, Ui_MainWindow):
         now = datetime.now()
         return now
 
-    def build_time_series(self) -> None:
+    def build_time_series(self, series: list) -> None:
         """
         Building time series from the collected
-        GPS data, HR data and current time.
+        GPS data, HR data and current time.\n
+        Args:
+            series (list[SensorData]):
+                list of series
         """
         HR = self.return_current_hr()
         GPS_LON, GPS_LAT, GPS_ALT = self.return_current_gps()
-        TIME = self.get_current_time()
-        DIST = self.calculate_distance()
-        self.time_series.append(
+        TIME = self.get_current_timestamp()
+        DIST = self.calculate_distance(series)
+        series.append(
             SensorData(HR, GPS_LON, GPS_LAT, GPS_ALT, TIME, DIST)
         )
 
@@ -295,7 +401,6 @@ class AST(QtWidgets.QMainWindow, Ui_MainWindow):
                 array.append(int(line.rstrip()))
 
             # Use data for calculation of average HR.
-            self.calculate_average_heart_rate(array)
             final = str(array[-1])
 
         return int(final)
@@ -321,36 +426,26 @@ class AST(QtWidgets.QMainWindow, Ui_MainWindow):
 
         return float(final[0]), float(final[1]), float(final[2])
 
-    def calculate_average_heart_rate(self, hr_data: list) -> float:
+    def calculate_distance(self, series: list) -> float:
         """
-        Calculate average HR of the workout.\n
+        Calculate distance of the workout.\n
         Args:
-            hr_data (list):
-                a list of all HR values of the workout
-        Returns:
-            float: average heart rate of the workout
-        """
-        av_hr = sum(hr_data) / len(hr_data)
-        return av_hr
-
-    def calculate_distance(self) -> float:
-        """
-        Calculate distance of workout.\n
+            series (list[SensorData])
         Returns:
             float: total distance
         """
-        if (len(self.time_series) < 2):
+        if (len(series) < 2):
             return 0.0
 
         total_distance = 0.0
-        for i in range(len(self.time_series) - 1):
+        for i in range(len(series) - 1):
             coords_1 = (
-                self.time_series[i + 1].longitude,
-                self.time_series[i + 1].latitude
+                series[i + 1].longitude,
+                series[i + 1].latitude
             )
             coords_2 = (
-                self.time_series[i].longitude,
-                self.time_series[i].latitude
+                series[i].longitude,
+                series[i].latitude
             )
             total_distance = (
                 total_distance +
@@ -432,11 +527,47 @@ class AST(QtWidgets.QMainWindow, Ui_MainWindow):
         self.stackedWidget.setCurrentIndex(index + 1)
         self.widget_title.setCurrentIndex(index + 1)
 
-    def start_interval(self) -> None:
+    def start_training(self) -> None:
         """
-        Manual interval start.
+        Manual training start.
         """
-        self.widget_interval_button.setCurrentIndex(1)
+        i = self.current_interval
+        if i >= len(self.planned_intervals):
+            return
+
+        self.start_interval_speed()
+
+        self.training_timer = QTimer()
+        self.training_timer.timeout.connect(self.end_interval_speed)
+        self.training_timer.start(
+            1000 * self.planned_intervals[i].speed_duration
+        )
+
+        self.current_interval += 1
+
+    def start_interval_speed(self):
+        """
+        Starting a speed part of an interval.
+        """
+        self.interval_tracker.start()
+        self.interval_flag = True
+        self.interval_average_heart_rate = 0
+        self.interval_duration = 0
+        self.interval_time = 0.0
+        self.is_speed_phase = True
+        self.lbl_interval_proposed_heart_rate.setText('-')
+
+        # Launching the digital twin.
+        interval = self.planned_intervals[self.current_interval]
+        self.digital_twin = DigitalTwin(
+            interval.speed_heart_rate,
+            interval.speed_duration,
+            TICK_TIME
+        )
+        self.digital_twin_thread = Thread(
+            target=self.digital_twin.predict_heart_rate
+        )
+        self.digital_twin_thread.start()
 
         # Show simple notification that an interval just started.
         self._feedback = TextFeedback(text='Interval started!')
@@ -446,20 +577,62 @@ class AST(QtWidgets.QMainWindow, Ui_MainWindow):
             time=3000
         )
 
-    def end_interval(self) -> None:
+    def end_interval_speed(self):
         """
-        Manual interval end.
+        Ending a speed part of an interval.
         """
-        self.widget_interval_button.setCurrentIndex(0)
+        self.interval_time = 0.0
+        self.interval_average_heart_rate = -1
+        self.interval_checkpoints = 0
+        self.lbl_interval_average_heart_rate.setText('-')
+        self.lbl_interval_proposed_heart_rate.setText('-')
+        self.is_speed_phase = False
 
-        # Show simple notification that an interval just ended.
-        self._feedback = TextFeedback(text='Interval ended!')
+        self.start_interval_rest()
+
+    def start_interval_rest(self):
+        """
+        Starting a rest part of an interval.
+        """
+        self.lbl_interval_watch.setText('0:00:00')
+        self.lbl_interval_proposed_heart_rate.setText('-')
+
+        # Show simple notification that an interval just started.
+        self._feedback = TextFeedback(text='Rest started!')
         self._feedback.show(
             AnimationType.HORIZONTAL,
-            AnimationDirection.LEFT,
+            AnimationDirection.RIGHT,
             time=3000
         )
 
+        interval = self.planned_intervals[self.current_interval]
+        self.training_timer = QTimer()
+        self.training_timer.timeout.connect(self.end_interval_rest)
+        self.training_timer.start(
+            1000 * interval.recovery_duration
+        )
+
+        # Launching the digital twin.
+        del self.digital_twin
+        self.digital_twin = DigitalTwin(
+            interval.recovery_heart_rate,
+            interval.recovery_duration,
+            TICK_TIME
+        )
+        self.digital_twin_thread = Thread(
+            target=self.digital_twin.predict_heart_rate
+        )
+        self.digital_twin_thread.start()
+
+    def end_interval_rest(self):
+        """
+        Ending a rest part of an interval.
+        """
+        self.lbl_interval_proposed_heart_rate.setText('-')
+        self.training_timer.stop()
+
+        self.start_training()
+
     def render_intervals_GUI(self) -> None:
         """
         Rendering the read intervals in the GUI.
@@ -474,7 +647,7 @@ class AST(QtWidgets.QMainWindow, Ui_MainWindow):
         # Inserting the read intervals.
         self.lyt_training.insertLayout(
             0,
-            Interval.render_intervals(self.intervals)
+            Interval.render_intervals(self.planned_intervals)
         )
 
     def load_training(self) -> None:
@@ -502,5 +675,16 @@ class AST(QtWidgets.QMainWindow, Ui_MainWindow):
         )
 
         # Adding and rendering the intervals from the training.
-        self.intervals = list(map(lambda i: i, training['interval']))
+        self.planned_intervals = []
+        for interval_group in training['interval']:
+            for i in range(interval_group['repetitions']):
+                interval = Interval(
+                    interval_group['speed_duration'],
+                    interval_group['recovery_duration'],
+                    interval_group['speed_heart_rate'],
+                    interval_group['recovery_heart_rate']
+                )
+                self.planned_intervals.append(interval)
+
+        # self.intervals = list(map(lambda i: i, training['interval']))
         self.render_intervals_GUI()
